# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
""" "Extras for exporting a captain agent."""

import json
import os
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

from waldiez.models import (
    WaldiezAgent,
    WaldiezCaptainAgent,
    WaldiezModel,
    WaldiezModelData,
)


def get_captain_agent_extras(
    agent: WaldiezAgent,
    agent_names: Dict[str, str],
    all_models: List[WaldiezModel],
    serializer: Callable[..., str],
    output_dir: Optional[Union[str, Path]],
) -> str:
    """Get the extra args for the captain agent.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    agent_names : Dict[str, str]
        A mapping of agent ids to agent names.
    all_models : List[WaldiezModel]
        All the models in the flow.
    serializer : Callable[..., str]
        The serializer to use.
    output_dir : Optional[Union[str, Path]]
        The output directory to save the agent lib and nested config.
    Returns
    -------
    str
        The extra args to use in the captain agent.
    """
    # extra args: nested_config, agent_lib, tool_lib
    if not isinstance(agent, WaldiezCaptainAgent):
        return ""
    agent_name = agent_names[agent.id]
    save_path = str(output_dir) if output_dir else "."
    extra_args_content = "\n" + "    agent_config_save_path=os.getcwd(),"
    if agent.data.agent_lib:
        lib_dict = [
            lib.model_dump(by_alias=False) for lib in agent.data.agent_lib
        ]
        lib_json_name = f"{agent_name}_agent_lib.json"
        agent_lib_path = os.path.join(save_path, lib_json_name)
        with open(agent_lib_path, "w", encoding="utf-8", newline="\n") as f:
            json.dump(lib_dict, f, ensure_ascii=False, indent=4)
        extra_args_content += "\n" + f'    agent_lib="{lib_json_name}",'
    if agent.data.tool_lib:
        extra_args_content += "\n" + f'    tool_lib="{agent.data.tool_lib}",'
    nested_config = generate_nested_config(
        agent,
        agent_name,
        all_models,
        save_path,
    )
    serialized_nested_config = serializer(nested_config)
    extra_args_content += (
        "\n" + f"    nested_config={serialized_nested_config},"
    )
    return extra_args_content


def generate_nested_config(
    agent: WaldiezCaptainAgent,
    agent_name: str,
    all_models: List[WaldiezModel],
    save_path: str,
) -> Dict[str, Any]:
    """Generate the nested config for the captain agent.

    Parameters
    ----------
    agent : WaldiezCaptainAgent
        The captain agent.
    agent_name : str
        The agent name.
    all_models : List[WaldiezModel]
        All the models in the flow.
    save_path : str
        The path to save the nested config.
    Returns
    -------
    Dict[str, Any]
        The nested config.
    """
    config_file_or_env_name = f"{agent_name}_llm_config.json"
    llm_config = get_llm_config(agent, all_models)
    os.makedirs(save_path, exist_ok=True)
    config_file_or_env_path = os.path.join(save_path, config_file_or_env_name)
    with open(
        config_file_or_env_path, "w", encoding="utf-8", newline="\n"
    ) as f:
        json.dump(llm_config, f, ensure_ascii=False, indent=4)
    nested_config = {
        "autobuild_init_config": {
            "config_file_or_env": config_file_or_env_name,
            "builder_model": llm_config["config_list"][0]["model"],
            "agent_model": llm_config["config_list"][0]["model"],
        },
        "autobuild_build_config": get_auto_build_build_config(
            agent, llm_config
        ),
        "group_chat_config": {"max_round": agent.data.max_round},
        "group_chat_llm_config": None,
        "max_turns": agent.data.max_turns,
    }
    return nested_config


def get_llm_config(
    agent: WaldiezAgent,
    all_models: List[WaldiezModel],
) -> Dict[str, Any]:
    """Get the config list environment variable name and its dict value.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    all_models : List[WaldiezModel]
        All the models in the flow.
    Returns
    -------
    Dict[str, str]
        The llm config dict.
    """
    model_name = "gpt-4o"
    temperature: Optional[float] = 1
    top_p: Optional[float] = 0.95
    max_tokens: Optional[int] = 2048
    config_dict: Dict[str, Any] = {}
    if agent.data.model_ids:
        waldiez_model = get_waldiez_model(agent.data.model_ids[0], all_models)
        llm_config = waldiez_model.get_llm_config(skip_price=True)
        for key in ["temperature", "top_p", "max_tokens"]:
            if key not in llm_config:
                llm_config[key] = None
        temp = llm_config.pop("temperature", None)
        config_dict = {
            "config_list": [llm_config],
        }
        if temp is not None:
            config_dict["temperature"] = temp
            return config_dict
    config_dict = {
        "model": model_name,
        "top_p": top_p,
        "max_tokens": max_tokens,
    }
    return {
        "config_list": [config_dict],
        "temperature": temperature,
    }


def get_auto_build_build_config(
    agent: WaldiezAgent,
    llm_config: Dict[str, Any],
) -> Dict[str, Any]:
    """Get the auto build build config.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    llm_config : Dict[str, Any]
        The llm config.

    Returns
    -------
    Dict[str, Any]
        The auto build build config.
    """
    coding = False
    code_execution_config = {
        "timeout": 300,
        "work_dir": "groupchat",
        "last_n_messages": 1,
        "use_docker": False,
    }
    if agent.data.code_execution_config is not False:
        coding = True
        code_execution_config["work_dir"] = (
            agent.data.code_execution_config.work_dir or "groupchat"
        )
        code_execution_config["last_n_messages"] = (
            agent.data.code_execution_config.last_n_messages or 1
        )
        code_execution_config["timeout"] = (
            agent.data.code_execution_config.timeout or 300
        )
    return {
        "default_llm_config": {
            "temperature": llm_config["temperature"],
            "top_p": llm_config["config_list"][0]["top_p"],
            "max_tokens": llm_config["config_list"][0]["max_tokens"],
        },
        "code_execution_config": code_execution_config,
        "coding": coding,
    }


def get_waldiez_model(
    model_id: str, all_models: List[WaldiezModel]
) -> WaldiezModel:
    """Get the model name from the model id.

    Parameters
    ----------
    model_id : str
        The model id.
    all_models : List[WaldiezModel]
        All the models in the flow.

    Returns
    -------
    str
        The model name.
    """
    for model in all_models:
        if model.id == model_id:
            return model
    now = (
        datetime.now(tz=timezone.utc)
        .isoformat(timespec="milliseconds")
        .replace("+00:00", "Z")
    )
    return WaldiezModel(
        id=model_id,
        type="model",
        name="gpt-4o",
        description="The GPT-4o model.",
        tags=["gpt-4o"],
        requirements=[],
        created_at=now,
        updated_at=now,
        data=WaldiezModelData(
            api_type="openai",
            temperature=1,
            top_p=0.95,
            max_tokens=2048,
            base_url=None,
            api_key=os.environ.get("OPENAI_API_KEY", "REPLACE_ME"),
            api_version=None,
            default_headers={},
            price=None,
        ),
    )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Get the standard imports for the flow exporter."""

from typing import List, Optional, Tuple

from waldiez.exporting.base import ImportPosition

BUILTIN_IMPORTS = [
    "import csv",
    "import json",
    "import os",
    "import sqlite3",
    "from dataclasses import asdict",
    "from pprint import pprint",
]
TYPING_IMPORTS = [
    "Annotated",
    "Any",
    "Callable",
    "Dict",
    "List",
    "Literal",
    "Optional",
    "Set",
    "Tuple",
    "Union",
]
COMMON_AUTOGEN_IMPORTS = [
    "from autogen import Agent",
    "from autogen import Cache",
    "from autogen import ConversableAgent",
    "from autogen import ChatResult",
    "from autogen import GroupChat",
    "from autogen import runtime_logging",
]


def sort_imports(
    all_imports: List[Tuple[str, ImportPosition]],
) -> Tuple[List[str], List[str], List[str], List[str], bool]:
    """Sort the imports.

    Parameters
    ----------
    all_imports : List[Tuple[str, ImportPosition]]
        All the imports.

    Returns
    -------
    Tuple[List[str], List[str], List[str], List[str], bool]
        The sorted imports and a flag if we got `import autogen`.
    """
    builtin_imports: List[str] = []
    third_party_imports: List[str] = []
    local_imports: List[str] = []
    autogen_imports: List[str] = COMMON_AUTOGEN_IMPORTS.copy()
    got_import_autogen = False
    for import_string, position in all_imports:
        if "import autogen" in import_string:
            got_import_autogen = True
            continue
        if import_string.startswith("from autogen"):
            autogen_imports.append(import_string)
            continue
        if position == ImportPosition.BUILTINS:
            builtin_imports.append(import_string)
        elif position == ImportPosition.THIRD_PARTY:
            third_party_imports.append(import_string)
        elif position == ImportPosition.LOCAL:
            local_imports.append(import_string)
    autogen_imports = list(set(autogen_imports))
    third_party_imports = ensure_np_import(third_party_imports)
    sorted_builtins = sorted(
        [imp for imp in builtin_imports if imp.startswith("import ")]
    ) + sorted([imp for imp in builtin_imports if imp.startswith("from ")])
    sorted_third_party = sorted(
        [imp for imp in third_party_imports if imp.startswith("import ")]
    ) + sorted([imp for imp in third_party_imports if imp.startswith("from ")])
    sorted_locals = sorted(
        [imp for imp in local_imports if imp.startswith("import ")]
    ) + sorted([imp for imp in local_imports if imp.startswith("from ")])

    return (
        sorted_builtins,
        sorted(autogen_imports),
        sorted_third_party,
        sorted_locals,
        got_import_autogen,
    )


def get_the_imports_string(
    all_imports: List[Tuple[str, ImportPosition]],
    is_async: bool,
) -> str:
    """Get the final imports string.

    Parameters
    ----------
    all_imports : List[Tuple[str, ImportPosition]]
        All the imports.
    is_async : bool
        If the flow is async.
    Returns
    -------
    str
        The final imports string.
    """
    (
        builtin_imports,
        autogen_imports,
        third_party_imports,
        local_imports,
        got_import_autogen,
    ) = sort_imports(all_imports)
    # Get the final imports string.
    # making sure, there are two lines after each import section
    # (builtin, third party, local)
    final_string = "\n".join(builtin_imports) + "\n"
    while not final_string.endswith("\n\n"):
        final_string += "\n"
    if is_async:
        final_string += (
            "\nimport aiofiles"
            "\nimport aiosqlite"
            "\nimport anyio"
            "\nimport nest_asyncio"
            "\nfrom aiocsv import AsyncDictWriter"
        )
    if got_import_autogen:
        final_string += "\nimport autogen  # type: ignore\n"
    if autogen_imports:
        final_string += "\n".join(autogen_imports) + "\n"
    if third_party_imports:
        final_string += "\n".join(third_party_imports) + "\n"
    while not final_string.endswith("\n\n"):
        final_string += "\n"
    if local_imports:
        final_string += "\n".join(local_imports) + "\n"
    while not final_string.endswith("\n\n"):
        final_string += "\n"
    if is_async:
        final_string += "\nnest_asyncio.apply()\n"
    return final_string.replace("\n\n\n", "\n\n")  # avoid too many newlines


def ensure_np_import(third_party_imports: List[str]) -> List[str]:
    """Ensure numpy is imported.

    Parameters
    ----------
    third_party_imports : List[str]
        The third party imports.

    Returns
    -------
    List[str]
        The third party imports with numpy.
    """
    if (
        not third_party_imports
        or "import numpy as np" not in third_party_imports
    ):
        third_party_imports.append("import numpy as np")
    return third_party_imports


def gather_imports(
    model_imports: Optional[List[Tuple[str, ImportPosition]]],
    skill_imports: Optional[List[Tuple[str, ImportPosition]]],
    chat_imports: Optional[List[Tuple[str, ImportPosition]]],
    agent_imports: Optional[List[Tuple[str, ImportPosition]]],
) -> List[Tuple[str, ImportPosition]]:
    """Gather all the imports.

    Parameters
    ----------
    model_imports : Tuple[str, ImportPosition]
        The model imports.
    skill_imports : Tuple[str, ImportPosition]
        The skill imports.
    chat_imports : Tuple[str, ImportPosition]
        The chat imports.
    agent_imports : Tuple[str, ImportPosition]
        The agent imports.

    Returns
    -------
    Tuple[str, ImportPosition]
        The gathered imports.
    """
    all_imports: List[Tuple[str, ImportPosition]] = []
    for import_statement in BUILTIN_IMPORTS:
        all_imports.append(
            (
                import_statement,
                ImportPosition.BUILTINS,
            )
        )
    if model_imports:
        all_imports.extend(model_imports)
    if skill_imports:
        all_imports.extend(skill_imports)
    if chat_imports:
        all_imports.extend(chat_imports)
    if agent_imports:
        all_imports.extend(agent_imports)
    # let's try to avoid this:
    # from typing import Annotated
    # from typing import Annotated, Any, Callable, Dict, ...Union
    all_typing_imports = TYPING_IMPORTS.copy()
    final_imports: List[Tuple[str, ImportPosition]] = []
    for import_statement, import_position in all_imports:
        if import_statement.startswith("from typing"):
            to_import = import_statement.split("import")[1].strip()
            if to_import:
                all_typing_imports.append(to_import)
        else:
            final_imports.append((import_statement, import_position))
    unique_typing_imports = list(set(all_typing_imports))
    one_typing_import = "from typing import " + ", ".join(
        sorted(unique_typing_imports)
    )
    final_imports.insert(1, (one_typing_import, ImportPosition.BUILTINS))
    return list(set(final_imports))



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez skill extra requirements."""

from typing import Iterator, Set

from .skill import WaldiezSkill


def get_skills_extra_requirements(
    skills: Iterator[WaldiezSkill],
    autogen_version: str,
) -> Set[str]:
    """Get the skills extra requirements.

    Parameters
    ----------
    skills : List[WaldiezSkill]
        The skills.
    autogen_version : str
        The ag2 version.
    Returns
    -------
    List[str]
        The skills extra requirements.
    """
    skill_requirements: Set[str] = set()
    for skill in skills:
        if skill.skill_type == "langchain":
            skill_requirements.add(
                f"pyautogen[interop-langchain]=={autogen_version}"
            )
        if skill.skill_type == "crewai":
            skill_requirements.add(
                f"pyautogen[interop-crewai]=={autogen_version}"
            )
        for requirement in skill.requirements:
            skill_requirements.add(requirement)
    return skill_requirements



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Function related utilities."""

import ast
import importlib.util
import sys
import sysconfig
from pathlib import Path
from typing import List, Optional, Tuple

import parso
import parso.python
import parso.tree

MAX_VAR_NAME_LENGTH = 64


def is_standard_library(module_name: str) -> bool:
    """Check if the module is part of the standard library.

    Parameters
    ----------
    module_name : str
        The module name.

    Returns
    -------
    bool
        True if the module is part of the standard library.
    """
    if module_name in sys.builtin_module_names:
        return True
    try:
        spec = importlib.util.find_spec(module_name)
    except BaseException:  # pylint: disable=broad-except
        return False
    if spec is None or not spec.origin:
        return False
    if "site-packages" in spec.origin:
        return False
    if spec.origin.startswith(sys.prefix) or spec.origin == "frozen":
        return True
    stdlib_path = str(Path(sysconfig.get_path("stdlib")).resolve())
    return spec.origin.startswith(stdlib_path)


def parse_code_string(
    code_string: str,
) -> Tuple[Optional[str], Optional[ast.Module]]:
    """Parse the code string.

    Parameters
    ----------
    code_string : str
        The code string.

    Returns
    -------
    Tuple[Optional[str], Optional[ast.Module]]
        If valid, None and the ast module.
        If invalid, the error message and None.
    """
    # pylint: disable=broad-except
    try:
        tree = ast.parse(code_string)
    except SyntaxError as e:
        return f"SyntaxError: {e}, in " + "\n" + f"{code_string}", None
    except BaseException as e:  # pragma: no cover
        return f"Invalid code: {e}, in " + "\n" + f"{code_string}", None
    return None, tree


def gather_code_imports(
    code_string: str,
    is_interop: bool,
) -> Tuple[List[str], List[str]]:
    """Gather the imports from the code string.

    Parameters
    ----------
    code_string : str
        The code string.
    is_interop : bool
        If True, make sure the interoperability import is present.

    Returns
    -------
    Tuple[List[str], List[str]]
        The standard library imports and the third party imports.
    """
    standard_lib_imports: List[str] = []
    third_party_imports: List[str] = []
    tree = parso.parse(code_string)  # type: ignore
    for node in tree.iter_imports():
        if node.type == "import_name":
            full_import_statement = node.get_code().strip()
            module_name = (
                node.get_code().replace("import", "").strip().split(" ")[0]
            )
            if not module_name:
                continue
            if is_standard_library(module_name):
                standard_lib_imports.append(full_import_statement)
            else:
                third_party_imports.append(full_import_statement)
        elif node.type == "import_from":
            full_import_statement = node.get_code().strip()
            module_name = (
                node.get_code().replace("from", "").strip().split(" ")[0]
            )
            if not module_name:
                continue
            if is_standard_library(module_name):
                standard_lib_imports.append(full_import_statement)
            else:
                third_party_imports.append(full_import_statement)
    if is_interop and (
        "from autogen.interop import Interoperability"
        not in third_party_imports
    ):
        third_party_imports.append(
            "from autogen.interop import Interoperability"
        )
    # sorted_standard_lib_imports =  # first import x, then from a import b
    sorted_standard_lib_imports = sorted(
        [stmt for stmt in standard_lib_imports if stmt.startswith("import ")]
    ) + sorted(
        [stmt for stmt in standard_lib_imports if stmt.startswith("from ")]
    )
    sorted_third_party_imports = sorted(
        [stmt for stmt in third_party_imports if stmt.startswith("import ")]
    ) + sorted(
        [stmt for stmt in third_party_imports if stmt.startswith("from ")]
    )
    return sorted_standard_lib_imports, sorted_third_party_imports


def check_function(
    code_string: str,
    function_name: str,
    function_args: List[str],
) -> Tuple[bool, str]:
    """Check the function.

    Parameters
    ----------
    code_string : str
        The code string to check.
    function_name : str
        The expected method name.
    function_args : List[str]
        The expected method arguments.
    Returns
    -------
    Tuple[bool, str]
        If valid, True and the function body (only), no extra lines.
        If invalid, False and the error message.
    """
    error, tree = parse_code_string(code_string)
    if error is not None or tree is None:
        return False, error or "Invalid code"
    return _validate_function_body(
        tree,
        code_string,
        function_name,
        function_args,
    )


def _validate_function_body(
    tree: ast.Module,
    code_string: str,
    function_name: str,
    function_args: List[str],
) -> Tuple[bool, str]:
    """Get the function body.

    Parameters
    ----------
    tree : ast.Module
        The ast module.
    function_body : str
        The function body.
    function_name : str
        The expected method name.
    function_args : List[str]
        The expected method arguments.
    Returns
    -------
    Tuple[bool, str]
        If valid, True and the function body (only), no extra lines.
        If invalid, False and the error message.
    """
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            if node.name != function_name:
                continue
            if len(node.args.args) != len(function_args):
                return (
                    False,
                    (
                        f"Invalid number of arguments, in function {node.name},"
                        f" expected: {len(function_args)},"
                        f" got: {len(node.args.args)} :("
                    ),
                )
            for arg, expected_arg in zip(node.args.args, function_args):
                if arg.arg != expected_arg:
                    return (
                        False,
                        (
                            f"Invalid argument name: {arg.arg}"
                            f" in function {node.name}"
                        ),
                    )
            if not node.body:  # pragma: no cover
                return False, "No body found in the function"
            function_body = _get_function_body(code_string, node)
            return True, function_body
    error_msg = (
        f"No method with name `{function_name}`"
        f" and arguments `{function_args}` found"
    )
    return False, error_msg


def get_function(
    code_string: str,
    function_name: str,
) -> str:
    """Get the function signature and body.

    Parameters
    ----------
    code_string : str
        The code string.
    function_name : str
        The function name.

    Returns
    -------
    str
        The function signature and body.
    """
    tree = parso.parse(code_string)  # type: ignore
    for node in tree.iter_funcdefs():
        if node.name.value == function_name:
            return node.get_code()
    return ""


def _get_function_body(
    code_string: str,
    node: ast.FunctionDef,
) -> str:
    """Get the function body, including docstring and comments inside.

    Parameters
    ----------
    code_string : str
        The code string.
    node : ast.FunctionDef
        The function node.

    Returns
    -------
    str
        The function body.

    Raises
    ------
    ValueError
        If no body found in the function.
    """
    lines = code_string.splitlines()
    signature_start_line = node.lineno - 1
    body_start_line = node.body[0].lineno - 1
    signature_end_line = signature_start_line
    for i in range(signature_start_line, body_start_line):
        if ")" in lines[i]:
            signature_end_line = i
            break
    function_body_lines = lines[signature_end_line + 1 :]
    last_line = function_body_lines[-1]
    if not last_line.strip() and len(function_body_lines) > 1:
        function_body_lines = function_body_lines[:-1]
    function_body = "\n".join(function_body_lines)
    while function_body.startswith("\n"):
        function_body = function_body[1:]
    while function_body.endswith("\n"):
        function_body = function_body[:-1]
    return function_body


def generate_function(
    function_name: str,
    function_args: List[str],
    function_types: Tuple[List[str], str],
    function_body: str,
    types_as_comments: bool = False,
) -> str:
    """Generate a function.

    Parameters
    ----------
    function_name : str
        The function name.
    function_args : List[str]
        The function arguments.
    function_types : Tuple[List[str], str]
        The function types.
    function_body : str
        The function body.
    types_as_comments : bool, optional
        Include the type hints as comments (or in the function signature)
        (default is False).
    Returns
    -------
    str
        The generated function.
    """
    if len(function_name) > MAX_VAR_NAME_LENGTH:
        function_name = function_name[:MAX_VAR_NAME_LENGTH]
    function_string = f"def {function_name}("
    if not function_args:
        function_string += ")"
    else:
        function_string += "\n"
        for arg, arg_type in zip(function_args, function_types[0]):
            if types_as_comments:
                function_string += f"    {arg},  # type: {arg_type}" + "\n"
            else:
                function_string += f"    {arg}: {arg_type}," + "\n"
        function_string += ")"
    if types_as_comments:
        function_string += ":\n"
        function_string += "    # type: (...) -> " + function_types[1]
    else:
        function_string += " -> " + function_types[1] + ":"
    function_string += "\n" if not function_body.startswith("\n") else ""
    function_string += f"{function_body}"
    if not function_string.endswith("\n"):
        function_string += "\n"
    return function_string
